

import java.util.Vector;
import java.awt.*;

class school { // implement swarming algorithm on school of fishes
  private Vector vFishes;
  static int minDISTANCE = 15; // separation distance 
  static int maxDISTANCE = 50; // detection radius

  school(int nBlue, int nRed) { // create vector of Blue and Red Fishs
    vFishes = new Vector(5, 5);
    for (int i=0; i < nBlue + nRed; i++)
      addFish(new fish((i < nBlue) ? Color.orange : Color.red));
    addFish(new catcher());
  }

  public void addFish(fish b) { // add Fish to vector
    vFishes.addElement(b);
  }

  synchronized void removeFish(Color col) { // remove Fish from vector
    for (int i=0; i < vFishes.size(); i++) { // loop through vector of Fishs
      fish bTemp = (fish)vFishes.elementAt(i);
      if (bTemp.getFins() == col) { // search for Fish of given colour
        vFishes.removeElementAt(i); // if found, remove Fish..
        break; // ..and stop searching
      }
    }
  }

/* The move function simply tells each Fish in the Vector vFishes to move 
according to the resultant Point of generalHeading */

  synchronized public void move() {
    for (int i=0; i < vFishes.size(); i++) {
      fish bTemp = (fish)vFishes.elementAt(i);
      bTemp.move(generalHeading(bTemp));
    }
  }

/* The generalHeading function determines the point a Fish will turn towards 
in the next timestep.  The Fish b checks for all Fishs (other than self) that 
fall within the detection range.  If the Fish is of a different colour or 
closer than the separation distance then they are repulsed else the Fishs are 
attracted. */

  synchronized private int generalHeading(fish b) {
    if (b instanceof hurdle && !(b instanceof catcher)) return 0;
    Point target = new Point(0, 0);
    int numFishs=0; // total of Fishs to average
    if (b instanceof catcher) {
      int distance = b.getDistance((fish)vFishes.elementAt(0));
      int index = 0;
      for (int i=1; i < vFishes.size(); i++) { // for each Fish in array
        fish bTemp = (fish)vFishes.elementAt(i); // retrieve element i
        int d = b.getDistance(bTemp);
        if (d > 0 && d < distance) {
          index = i;
          distance = d;
        }
      }
      fish targetFish = (fish)vFishes.elementAt(index);
      target.x = targetFish.getX();
      target.y = targetFish.getY();
      if (distance <= 10) { // if Chaser in contact range
        int ChaserIndex = vFishes.indexOf(b);
        fish newFish = new fish(b.getX(), b.getY(), (b.getTheta() + 180) % 360, b.getFins());
        vFishes.setElementAt(newFish, index);
        catcher newChaser = new catcher(targetFish.getX(), targetFish.getY(), targetFish.getTheta(), targetFish.getFins());
        vFishes.setElementAt(newChaser, ChaserIndex);
      }
    } else {
      for (int i=0; i < vFishes.size(); i++) { // for each Fish in array
        fish bTemp = (fish)vFishes.elementAt(i); // retrieve element i
        int distance = b.getDistance(bTemp); // get distance to Fish

        if (!b.equals(bTemp) && distance < maxDISTANCE && distance > 0) {

/* In attraction the target point moves toward the aim point of the 
other Fish unless it is too close in which case the Fishs aim to align their 
directions of travel. */

          int xdist = 0; int ydist = 0;
          if (b.getFins().equals(bTemp.getFins()) && !(bTemp instanceof catcher)) {
            xdist = bTemp.getAim().x - bTemp.getX(); // alignment component
            ydist = bTemp.getAim().y - bTemp.getY();
            if (distance < minDISTANCE) {
              xdist -= bTemp.getAim().x - b.getX(); // repulsion component
              ydist -= bTemp.getAim().y - b.getY();
            } else {
              xdist += bTemp.getAim().x - b.getX(); // attraction component
              ydist += bTemp.getAim().y - b.getY();
            }
            double pdist = Math.sqrt((xdist)*(xdist) + (ydist)*(ydist));
            target.x += xdist * 100/pdist; // normalise
            target.y += ydist * 100/pdist;
          }

/* In repulsion the target point moves away from the other Fish with a 
weight over attraction of 10 scaled by (1-distance/maxDISTANCE) which is 
maximised when distance is zero and minimised at the perimeter of the 
detection area -- the repulsion decreases linearly. */

          else { // repulsion
            xdist = b.getX() - bTemp.getX();
            ydist = b.getY() - bTemp.getY();
            double pdist = Math.sqrt((xdist)*(xdist) + (ydist)*(ydist));
            target.x += xdist * 1000/pdist * (1-(double)distance/(double)maxDISTANCE);
            target.y += ydist * 1000/pdist * (1-(double)distance/(double)maxDISTANCE);
          }
          numFishs++;
        }
      }
      if (numFishs == 0) { // keep going straight
        target.x = b.getAim().x;
        target.y = b.getAim().y;
      } else { // average points to find target position
        target.x = b.getX() + target.x/numFishs;
        target.y = b.getY() + target.y/numFishs;
      }
    }
    double grad = (double)(b.getY() - target.y)/(double)(target.x - b.getX());
    int targetTheta = (int)(Math.atan(grad) * (180/Math.PI));
    if (target.x < b.getX()) targetTheta += 180;
    if (targetTheta < 0) targetTheta += 360;
    return targetTheta; // angle for Fish to steer towards
  }

  synchronized public void display(Graphics g) { // display each Fish in flock
    for (int i = 0; i < vFishes.size(); i++) {
      fish fTemp = (fish)vFishes.elementAt(i);
      fTemp.display(g);
    }
  }
}